#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0

//BG0 
//SPR t[

//---------------------------------------------------------------------------
#define STATE_INIT 				1		//is
#define STATE_TITLE				2
#define STATE_GAME				3
#define STATE_END				4

#define STAGE_FRAME_CNT			4		//t[
#define STAGE_WAIT_CNT			40		//EFCg
#define STAGE_RADIUS_X			60		//R[X̉̔
#define STAGE_RADIUS_Y			40		//R[X̏c̔
#define STAGE_MOVE_X			6		//XWړ
#define STAGE_MOVE_Y			4		//YWړ

#define STAGE_X_MID				60		//\̒SXW
#define STAGE_Y_MID				40		//\̒SYW
#define STAGE_Z_STEP			10		//t[Ԃ̋
#define STAGE_Z_MAX				50		//ZWől
#define STAGE_Z_MIN				10		//ZWŏl

//---------------------------------------------------------------------------
typedef struct {
	s16 x;
	s16 y;
	s16 z;
} ST_FRAME;

typedef struct {
	ST_FRAME frame[STAGE_FRAME_CNT];
	s16      sx;					//ړ
	s16      sy;
} ST_STAGE;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_BG    BG0;
ST_STAGE Stage;
u16      Score;
u16      Key;
u16      Wait;
u8       State;						//is󋵃tO
//---------------------------------------------------------------------------
void StateInit(); 
void StateTitle();
void StateGame(); 
void StateEnd();  

void InitWS();
void InitBG();
void InitKey();
void InitSprite();
void IRQUserHandler();

void SetStage();
void SetScore();

void DrawScore();
void DrawFrame();

void ProcKey();
void ScanKey();

void PlayerMove();
void StageMove();
void UpdateFrame();

bool IsCourseOut();

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_TITLE: StateTitle(); break;
			case STATE_GAME:  StateGame();  break;
			case STATE_END:   StateEnd();   break;

			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitBG();
	InitKey();
	InitSprite();

	State = STATE_TITLE;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG0 

	BG0.Chr = 1;
	BG0.Map = 0;
	BG0.Col = 0;

	REG_BG0CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
	              BG_CHARBASE(BG0.Chr) | BG_MAPBASE(BG0.Map) | 0);

	BGSetPal16  (BG0.Col, (u16*)&BGFont_palette);
	BGSetChrData(BG0.Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteInit();

	SpriteSetPal16((u16*)&spr_chr_pal_bin);
	SpriteSetData ((u16*)&spr_chr_img_bin, SPR_CHR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();
	Key = KEY_NONE;

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;
	REG_TM0D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void SetStage()
{
	ST_STAGE* p = &Stage;
	s16 i;

	//Frame
	for(i=0; i<STAGE_FRAME_CNT; i++)
	{
		p->frame[i].x = 0;
		p->frame[i].y = 0;
		p->frame[i].z = (i+1) * STAGE_Z_STEP + STAGE_Z_MIN - 1;
	}
	p->sx   = 0;
	p->sy   = 0;

	//Sprite
	for(i=0; i<4*4 ; i++)
	{
		SpriteSetSize(i, SP_SIZE_8, SP_SQUARE, SP_COLOR_16);
		SpriteSetChr (i, i);
	}
}
//---------------------------------------------------------------------------
void SetScore()
{
	Score = 0;
	Wait  = STAGE_WAIT_CNT;
}
//---------------------------------------------------------------------------
void StateTitle()
{
	SetStage();
	SetScore();

	BGDrawCls(BG0.Map, ' ');
	BGDrawStr(BG0.Map, 1, 2, (u8*)"Cosmic Air Way");
	BGDrawStr(BG0.Map, 4, 4, (u8*)"ver.1.00      ");
	BGDrawStr(BG0.Map, 3, 8, (u8*)"PUSH START    ");

	SetMode(MODE_0 | BG0_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	WaitKey(KEY_START);

	BGDrawCls(BG0.Map, ' ');
	State = STATE_GAME;
}
//---------------------------------------------------------------------------
void StateGame()
{
	WaitForVsync();
	DrawScore();
	DrawFrame();

	ProcKey();

	PlayerMove();
	StageMove();

	if( IsCourseOut() == TRUE )
	{
		State = STATE_END;
		return;
	}

	UpdateFrame();

	WaitMs(Wait);
}
//---------------------------------------------------------------------------
void DrawScore()
{
	char msg[64];
	bprintf(msg, "s:%4d  m:%4d", STAGE_WAIT_CNT - Wait, Score);
	BGDrawStr(BG0.Map, 0, 0, (u8*)msg);
}
//---------------------------------------------------------------------------
void ScanKey()
{
	if( KeyWaitTimer() == TRUE )
	{
		return;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return;
	}

	Key = KeyInput();
}
//---------------------------------------------------------------------------
void ProcKey()
{
	ScanKey();

	//|[Y`FbN܂
	if(Key & KEY_START)
	{
		BGDrawStr(BG0.Map, 5, 5, (u8*)"PAUSE");
		WaitKey2(KEY_START);
		BGDrawStr(BG0.Map, 5, 5, (u8*)"     ");
	}
}
//---------------------------------------------------------------------------
void PlayerMove()
{
	ST_STAGE* p = &Stage;
	s16 moveX, moveY; 
	s16 i;

	//ꂽ{^ňړʂ߂܂
	moveX  = Key & KEY_B ? STAGE_MOVE_X / 2 : STAGE_MOVE_X;
	moveY  = Key & KEY_A ? STAGE_MOVE_Y / 2 : STAGE_MOVE_Y;

	//ꂽ{^ňړ߂܂
	moveX *= Key & KEY_RIGHT ? -1 :  1;
	moveY *= Key & KEY_DOWN  ?  1 : -1;

	if( (Key & KEY_RIGHT) || (Key & KEY_LEFT) )
	{
		for(i=0; i<STAGE_FRAME_CNT; i++)
		{
			p->frame[i].x += moveX;
		}
	}

	if( (Key & KEY_UP) || (Key & KEY_DOWN) )
	{
		for(i=0; i<STAGE_FRAME_CNT; i++)
		{
			p->frame[i].y += moveY;
		}
	}
}
//---------------------------------------------------------------------------
void StageMove()
{
	ST_STAGE* p = &Stage;
	s16 i;

	for(i=0; i<STAGE_FRAME_CNT; i++)
	{
		p->frame[i].z--;
	}
}
//---------------------------------------------------------------------------
bool IsCourseOut()
{
	ST_STAGE* p = &Stage;
	s16 i;

	for(i=0; i<STAGE_FRAME_CNT; i++)
	{
		if(p->frame[i].z < STAGE_Z_MIN)
		{
			if(   (p->frame[i].x < -STAGE_RADIUS_X)
			   || (p->frame[i].x >  STAGE_RADIUS_X)
			   || (p->frame[i].y < -STAGE_RADIUS_Y)
			   || (p->frame[i].y >  STAGE_RADIUS_Y) )
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
void UpdateFrame()
{
	ST_STAGE* p = &Stage;
	s16 i;

	if(p->frame[0].z >= STAGE_Z_MIN)
	{
		return;
	}

	//t[ړ܂
	for(i=0; i<STAGE_FRAME_CNT-1; i++)
	{
		p->frame[i] = p->frame[i + 1];
	}
	p->frame[STAGE_FRAME_CNT - 1].x += p->sx;
	p->frame[STAGE_FRAME_CNT - 1].y += p->sy;
	p->frame[STAGE_FRAME_CNT - 1].z  = STAGE_Z_MAX - 1;

	if( (Score + STAGE_FRAME_CNT) % 10 == 0)
	{
		//t[̈ړʂ̕ύX܂
		u32 rnd = (u32)(rand() % (2 * STAGE_Z_STEP)) - STAGE_Z_STEP;
		p->sx   = STAGE_MOVE_X * rnd;
		p->sy   = STAGE_MOVE_Y * rnd;

		if(Wait > 0)
		{
			Wait--;
		}
	}

	Score++;
}
//---------------------------------------------------------------------------
#define ROUND_X(X, Z) (s16)((f32)STAGE_Z_MIN / Z * (X) + STAGE_X_MID + 0.5) //XWۂ
#define ROUND_Y(Y, Z) (s16)((f32)STAGE_Z_MIN / Z * (Y) + STAGE_Y_MID + 0.5) //YWۂ

void DrawFrame()
{
	ST_STAGE* p = &Stage;
	s16 x[4];	//4́uAEAAEv̕ӂ̐ł
	s16 y[4];
	s16 i;

	for(i=0; i<STAGE_FRAME_CNT; i++)
	{
		x[0] = ROUND_X(p->frame[i].x - STAGE_RADIUS_X, p->frame[i].z) + 4;
		y[0] = ROUND_Y(p->frame[i].y - STAGE_RADIUS_Y, p->frame[i].z) + 4;
		x[1] = ROUND_X(p->frame[i].x + STAGE_RADIUS_X, p->frame[i].z) + 4;
		y[1] = ROUND_Y(p->frame[i].y - STAGE_RADIUS_Y, p->frame[i].z) + 4;
		x[2] = ROUND_X(p->frame[i].x - STAGE_RADIUS_X, p->frame[i].z) + 4;
		y[2] = ROUND_Y(p->frame[i].y + STAGE_RADIUS_Y, p->frame[i].z) + 4;
		x[3] = ROUND_X(p->frame[i].x + STAGE_RADIUS_X, p->frame[i].z) + 4;
		y[3] = ROUND_Y(p->frame[i].y + STAGE_RADIUS_Y, p->frame[i].z) + 4;

		SpriteMove(i*4+0, x[0], y[0]);
		SpriteMove(i*4+1, x[1], y[1]);
		SpriteMove(i*4+2, x[2], y[2]);
		SpriteMove(i*4+3, x[3], y[3]);
	}
}
//---------------------------------------------------------------------------
void StateEnd()
{
	BGDrawStr(BG0.Map, 3, 5, (u8*)"COURSE OUT");
	WaitKey(KEY_START);

	State = STATE_TITLE;
}
